from oarepo_model_builder.model_preprocessors import ModelPreprocessor


class InvenioBaseClassesModelPreprocessor(ModelPreprocessor):
    TYPE = "invenio_base_classes"

    def transform(self, schema, settings):
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-resource-bases",
            [],
            "invenio_records_resources.resources.RecordResource",
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-resource-config-bases",
            [],
            "invenio_records_resources.resources.RecordResourceConfig",
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-service-bases",
            [],
            "invenio_records_resources.services.RecordService",
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-bases",
            [],
            "invenio_records_resources.records.api.Record",
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-service-config-bases",
            [],
            "invenio_records_resources.services.RecordServiceConfig",
        )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordPermissionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_permissions"
    class_config = "record-permissions-class"
    template = "record-permissions"



import re

from marshmallow import fields
from marshmallow.exceptions import ValidationError

from .datatypes import DataType


def validate_regex(value):
    if not value:
        return
    try:
        re.compile(value)
    except Exception as e:
        raise ValidationError(
            f"Regex '{value}' is not valid. Reported error: '{str(e)}'"
        )


class StringDataType(DataType):
    schema_type = "string"
    marshmallow_field = "ma_fields.String"

    class ModelSchema(DataType.ModelSchema):
        minLength = fields.Integer(required=False)
        maxLength = fields.Integer(required=False)
        regex = fields.String(required=False, validate=validate_regex)
        enum = fields.List(fields.String(), required=False)

    def marshmallow_validators(self):
        validators = []
        ranges = {}
        for param, schema in (
            ("min", "minLength"),
            ("max", "maxLength"),
        ):
            if schema in self.definition:
                ranges[param] = self.definition[schema]

        if ranges:
            params = ", ".join(f"{k}={v}" for k, v in ranges.items())
            validators.append(f"ma_validate.Length({params})")

        if "pattern" in self.definition:
            pattern = self.definition["pattern"]
            pattern = pattern.replace("\\", "\\\\")
            pattern = pattern.replace('"', '\\"')
            validators.append(f'ma_validate.Regexp("{pattern}")')

        return validators


class FulltextDataType(StringDataType):
    mapping_type = "text"
    model_type = "fulltext"


class KeywordDataType(StringDataType):
    mapping_type = "keyword"
    model_type = "keyword"

    def facet(self, nested_facet):
        return f"TermsFacet({self.path})"


class FulltextKeywordDataType(StringDataType):
    mapping_type = "text"
    model_type = "fulltext+keyword"

    def mapping(self):
        ret = super().mapping()
        mapping_el = ret.setdefault("mapping", {})
        mapping_el.setdefault("fields", {}).setdefault("keyword", {"type": "keyword"})
        return ret



"""
Datatypes supported out of the box:

integer, float, double:
  constraints: minimumExclusive, maximumExclusive, minimum, maximum

boolean:
  no constraints supported

date, time, datetime
  no constraints supported

edtf, edtf-interval
  no constraints supported

fulltext, keyword, fulltext+keyword:
  minLength, maxLength, pattern

flatten:
  object represented as flatten type in elasticsearch
"""


from .containers import (ArrayDataType, FlattenDataType, NestedDataType,
                         ObjectDataType)
from .datatypes import DataType, Import, datatypes  # noqa
from .dates import (DateDataType, DateTimeDataType, EDTFDataType,
                    EDTFIntervalType, TimeDataType)
from .primitive_types import NumberDataType  # noqa
from .primitive_types import (BooleanDataType, DoubleDataType, FloatDataType,
                              IntegerDataType)
from .strings import StringDataType  # noqa
from .strings import FulltextDataType, FulltextKeywordDataType, KeywordDataType

DEFAULT_DATATYPES = [
    IntegerDataType,
    FloatDataType,
    DoubleDataType,
    BooleanDataType,
    DateDataType,
    TimeDataType,
    DateTimeDataType,
    EDTFDataType,
    EDTFIntervalType,
    FulltextDataType,
    KeywordDataType,
    FulltextKeywordDataType,
    ObjectDataType,
    NestedDataType,
    FlattenDataType,
    ArrayDataType,
]



class ModelPreprocessor:
    def __init__(self, builder: "oarepo_model_builder.builder.ModelBuilder"):
        self.builder = builder

    def transform(self, schema, settings):
        pass

    def set(self, settings, name, callable=None):
        if callable:
            if not settings.get(name):
                settings[name] = callable()
            return

        def w(func):
            if not settings.get(name):
                settings[name] = func()

        return w

    def set_default_and_append_if_not_present(self, container, key, value, appended):
        container.setdefault(key, value)
        arr = getattr(container, key)
        if appended not in arr:
            arr.append(appended)

    def set_default_and_prepend_if_not_present(self, container, key, value, prepended):
        container.setdefault(key, value)
        arr = getattr(container, key)
        if prepended not in arr:
            arr.insert(0, prepended)



import lazy_object_proxy
from libcst import (Arg, Assign, Call, ClassDef, Dict, Element, Expr,
                    FunctionDef, Import, ImportFrom, Integer, List, Module,
                    Name, Pass, SimpleStatementLine, SimpleString,
                    StarredElement, Tuple)


@lazy_object_proxy.Proxy
def module_mergers():
    from .indented_nodes import ModuleMerger

    return {Module: ModuleMerger()}


@lazy_object_proxy.Proxy
def indented_block_mergers():
    from .indented_nodes import ClassMerger
    from .simple_nodes import FunctionMerger, SimpleStatementLineMerger

    return {
        SimpleStatementLine: SimpleStatementLineMerger(),
        ClassDef: ClassMerger(),
        FunctionDef: FunctionMerger(),
    }


@lazy_object_proxy.Proxy
def simple_line_mergers():
    from .simple_nodes import (AssignMerger, ExprMerger, ImportFromMerger,
                               ImportMerger, PassMerger)

    return {
        Assign: AssignMerger(),
        Import: ImportMerger(),
        ImportFrom: ImportFromMerger(),
        Expr: ExprMerger(),
        Pass: PassMerger(),
    }


@lazy_object_proxy.Proxy
def call_mergers():
    from .call import ArgMerger

    return {Arg: ArgMerger()}


@lazy_object_proxy.Proxy
def expression_mergers():
    from oarepo_model_builder.utils.cst.collections import (DictMerger,
                                                            ElementMerger,
                                                            ListMerger)

    from .call import CallMerger
    from .simple_nodes import (ExprMerger, IntegerMerger, NameMerger,
                               SimpleStringMerger, StarredElementMerger)

    return {
        Call: CallMerger(),
        List: ListMerger(),
        Tuple: ListMerger(),
        Integer: IntegerMerger(),
        Element: ElementMerger(),
        SimpleString: SimpleStringMerger(),
        Name: NameMerger(),
        Expr: ExprMerger(),
        Dict: DictMerger(),
        StarredElement: StarredElementMerger(),
    }



from __future__ import annotations

import functools
import inspect
from collections import namedtuple
from typing import TYPE_CHECKING

from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.json_pathlib import JSONPaths

if TYPE_CHECKING:
    from oarepo_model_builder.builder import ModelBuilder


class PropertyPreprocessor:
    PathMethodRecord = namedtuple("PathMethodRecord", "method, output_builder_type")

    def __init__(self, builder: ModelBuilder):
        self.builder = builder
        # TODO: move this to metaclass and initialize only once per class
        self.json_paths = JSONPaths()
        arr = []
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if not hasattr(method, "model_builder_priority"):
                continue
            for ot in method.model_builder_output_builder_types:
                arr.append(
                    (
                        -method.model_builder_priority,
                        -len(method.model_builder_path),
                        method.model_builder_path,
                        id(method),
                        method.model_builder_condition,
                        ot,
                        method,
                    )
                )
        arr.sort()
        for _prior, _lpath, path, _mid, condition, output_builder_type, method in arr:
            self.json_paths.register(
                path,
                condition,
                PropertyPreprocessor.PathMethodRecord(method, output_builder_type),
            )

    def begin(self, schema, settings):
        self.schema = schema
        self.settings = settings

    def finish(self):
        pass

    def _call_method(self, data, stack: ModelBuilderStack, output_builder_type):
        for method, _output_builder_type in self.json_paths.match(
            stack.path, stack.top.data, extra_data={"stack": stack}
        ):
            if (
                _output_builder_type == "*"
                or output_builder_type == _output_builder_type
            ):
                return method(
                    data, stack=stack, output_builder_type=output_builder_type
                )

    def process(self, output_builder_type: str, data, stack: ModelBuilderStack):
        return self._call_method(data, stack, output_builder_type)


def process(model_builder, path, priority=0, condition=None):
    """
    :param model_builder either a single builder, its TYPE property or a list of builders or their TYPEs
    :param path A single path or a tuple of paths that must be matched. May contain '**' which means 1+ path segments,
                * which means single path segment (for example properties/blah might is matched by **
                or properties/*)
    """

    def convert_type(model_builder_types):
        if not isinstance(model_builder_types, (list, tuple)):
            model_builder_types = [model_builder_types]
        return [
            model_builder if isinstance(model_builder, str) else model_builder.TYPE
            for model_builder in model_builder_types
        ]

    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        wrapped.model_builder_priority = priority
        wrapped.model_builder_output_builder_types = convert_type(model_builder)
        wrapped.model_builder_path = path
        wrapped.model_builder_condition = condition
        return wrapped

    return wrapper



import marshmallow as ma
from marshmallow import fields

from .model_validation import model_validator
from .utils import ExtendablePartSchema


class ImportSchema(ExtendablePartSchema):
    _import = fields.String(data_key="import", required=True)
    alias = fields.String(required=False)


class PropertyMarshmallowSchema(ExtendablePartSchema):
    read = fields.Boolean(required=False)
    write = fields.Boolean(required=False)
    imports = fields.List(fields.Nested(ImportSchema), required=False)
    field_name = fields.String(data_key="field-name", required=False)
    field = fields.String(required=False)
    field_class = fields.String(data_key="field-class", required=False)
    arguments = fields.List(fields.String(), required=False)
    validators = fields.List(fields.String(), required=False)


class ModelMarshmallowSchema(ma.Schema):
    class ObjectOnlyMarshmallowProps(ExtendablePartSchema):
        imports = fields.List(
            fields.Nested(ImportSchema), required=False
        )  # imports must be here as well as it is used on model's root (without field)
        generate = fields.Boolean(required=False)
        schema_class = fields.String(data_key="schema-class", required=False)
        base_classes = fields.List(
            fields.String(), data_key="base-classes", required=False
        )

    marshmallow = fields.Nested(
        lambda: model_validator.validator_class("property-marshmallow-model")()
    )


class ObjectPropertyMarshmallowSchema(ma.Schema):
    class ObjectMarshmallowProps(
        ModelMarshmallowSchema.ObjectOnlyMarshmallowProps,
        PropertyMarshmallowSchema,
    ):
        pass

    marshmallow = fields.Nested(
        lambda: model_validator.validator_class("property-marshmallow-object")()
    )



from oarepo_model_builder.outputs.json_stack import JSONStack


def test_json_stack():
    st = JSONStack()
    st.push(None, {})
    st.push("a", 1)
    st.pop()
    st.push("b", {})
    st.push("c", 2)
    st.pop()
    st.pop()
    st.push("d", [])
    st.push(0, 3)
    st.pop()
    st.push(1, 4)
    st.pop()
    st.pop()
    st.pop()
    st.pop()
    assert st.value == {"a": 1, "b": {"c": 2}, "d": [3, 4]}



from marshmallow import fields

from .model_validation import model_validator
from .utils import ExtendablePartSchema


class RootSchema(ExtendablePartSchema):
    model = fields.Nested(lambda: model_validator.validator_class("model")())
    settings = fields.Nested(lambda: model_validator.validator_class("settings")())
    version = fields.String(required=False, dump_default="1.0.0")
    title = fields.String(required=False)
    output_directory = fields.String(data_key="output-directory", required=False)



import json
import os
import re
from io import StringIO
from pathlib import Path
from typing import Dict

from oarepo_model_builder.entrypoints import (create_builder_from_entrypoints,
                                              load_model)
from oarepo_model_builder.fs import InMemoryFileSystem

DUMMY_YAML = "test.yaml"


def test_include_invenio():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "fulltext+keyword"},  # NOSONAR
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},  # NOSONAR
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl

class NestedLabeledFacet(dsl.Facet):
    agg_type = "nested"

    def __init__(self, path, nested_facet, label = ''):
        self._path = path
        self._inner = nested_facet
        self._label = label
        super(NestedLabeledFacet, self).__init__(
            path=path,  aggs={"inner": nested_facet.get_aggregation(),}
        )

    def get_values(self, data, filter_values):
        return self._inner.get_values(data.inner, filter_values)

    def add_filter(self, filter_values):
        inner_q = self._inner.add_filter(filter_values)
        if inner_q:
            return dsl.Nested(path=self._path, query=inner_q)

    def get_labelled_values(self, data, filter_values):
        \"""Get a labelled version of a bucket.\"""
        try:
            out = data['buckets']
        except:
            out = []
        return {'buckets': out, 'label': str(self._label)}



a_keyword = TermsFacet(field = "a.keyword")



b = TermsFacet(field="cosi")



_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
    """,
    )


def test_sort():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {
                        "type": "fulltext+keyword",
                        "sortable": {"key": "a_test"},
                    },
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                        "sortable": {"key": "b_test", "order": "desc"},
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "search.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services import SearchOptions as InvenioSearchOptions
from . import facets

def _(x):
    \"""Identity function for string extraction.\"""
    return x

class TestSearchOptions(InvenioSearchOptions):
    \"""TestRecord search options.\"""

    facets = {
    'a_keyword': facets.a_keyword,
    'b': facets.b,
    '_id': facets._id,
    'created': facets.created,
    'updated': facets.updated,
    '_schema': facets._schema,
    }
    sort_options = {
        **InvenioSearchOptions.sort_options,
    'a_test': {'fields': ['a']},
    'b_test': {'fields': ['-b']},
    }
    """,
    )


def test_nested():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b": {
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "properties": {"g": {"type": "keyword"}},
                                "mapping": {"type": "nested"},
                                "marshmallow": {
                                    "schema-class": "nest.f.F",
                                    "generate": True,
                                },
                            },
                        },
                        "mapping": {"type": "nested"},
                        "marshmallow": {"schema-class": "nest.b.B", "generate": True},
                    }
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_records_resources.services.records.facets import TermsFacet
from invenio_search.engine import dsl

class NestedLabeledFacet(dsl.Facet):
    agg_type = "nested"

    def __init__(self, path, nested_facet, label = ''):
        self._path = path
        self._inner = nested_facet
        self._label = label
        super(NestedLabeledFacet, self).__init__(
            path=path,  aggs={\"inner\": nested_facet.get_aggregation(),}
        )

    def get_values(self, data, filter_values):
        return self._inner.get_values(data.inner, filter_values)

    def add_filter(self, filter_values):
        inner_q = self._inner.add_filter(filter_values)
        if inner_q:
            return dsl.Nested(path=self._path, query=inner_q)

    def get_labelled_values(self, data, filter_values):
        \"""Get a labelled version of a bucket.\"""
        try:
            out = data['buckets']
        except:
            out = []
        return {'buckets': out, 'label': str(self._label)}



b_c = NestedLabeledFacet(path = \"b\" , nested_facet =TermsFacet(field = \"b.c\"))



b_d_keyword = NestedLabeledFacet(path = \"b\" , nested_facet =TermsFacet(field = \"b.d.keyword\"))



b_f_g = NestedLabeledFacet(path = \"b\", nested_facet = NestedLabeledFacet(path = "b.f" , nested_facet =TermsFacet(field = "b.f.g")))



_id = TermsFacet(field = "id")



created = TermsFacet(field = "created")



updated = TermsFacet(field = "updated")



_schema = TermsFacet(field = "$schema")
    """,
    )


def test_search_class():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "fulltext+keyword"},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "search.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services import SearchOptions as InvenioSearchOptions
from . import facets

def _(x):
    \"""Identity function for string extraction.\"""
    return x

class TestSearchOptions(InvenioSearchOptions):
    \"""TestRecord search options.\"""

    facets = {
      'a_keyword': facets.a_keyword,
      'b': facets.b,
      '_id': facets._id,
      'created': facets.created,
      'updated': facets.updated,
      '_schema': facets._schema,
    }
    sort_options = {
        **InvenioSearchOptions.sort_options,
    }
    """,
    )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_resource"
    class_config = "record-resource-class"
    template = "record-resource"



import json
import os
import re
import shutil
import sys
from pathlib import Path
from tempfile import mkdtemp, tempdir

from oarepo_model_builder.entrypoints import (create_builder_from_entrypoints,
                                              load_model)
from oarepo_model_builder.fs import FileSystem, InMemoryFileSystem
from tests.utils import assert_python_equals

OAREPO_USE = "use"


def test_empty_metadata():
    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "version": "1.0.0",
            "model": {OAREPO_USE: "invenio", "properties": None},
        },
        isort=False,
        black=False,
    )

    filesystem = FileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem, overwrite=True)

    tmpdir = mkdtemp()
    try:
        builder.build(schema, tmpdir)
    finally:
        shutil.rmtree(tmpdir)



import json
import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.model_saver import (
    ModelRegistrationBuilder, ModelSaverBuilder)
from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.entrypoints import (
    load_entry_points_dict, load_included_models_from_entry_points)
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.model_preprocessors.default_values import \
    DefaultValuesModelPreprocessor
from oarepo_model_builder.model_preprocessors.invenio import \
    InvenioModelPreprocessor
from oarepo_model_builder.outputs.cfg import CFGOutput
from oarepo_model_builder.outputs.json import JSONOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.property_preprocessors.datatype_preprocessor import \
    DataTypePreprocessor
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.validation.model_validation import model_validator
from tests.multilang import (MultilangPreprocessor, MultilingualDataType,
                             UIValidator)

try:
    import json5
except ImportError:
    import json as json5


def test_model_saver():
    data = build(
        {
            "properties": {
                "a": {"type": "keyword", "ui": {"class": "bolder"}},
                "b": {
                    "type": "object",
                    "properties": {
                        "c": {
                            "type": "keyword",
                        }
                    },
                    "marshmallow": {"generate": True},
                },
                "metadata": {"properties": {}},
            }
        },
        property_preprocessors=[
            DataTypePreprocessor,
        ],
    )
    print(json.dumps(data[0]))
    assert data[0] == {
        "model": {
            "type": "object",
            "record-dumper-class": "test.records.dumper.TestDumper",
            "invenio-record-dumper-extra-code": "",
            "schema-version": "1.0.0",
            "record-resource-class": "test.resources.records.resource.TestResource",
            "marshmallow": {
                "schema-class": "test.services.records.schema.TestSchema",  # NOSONAR
                "base-classes": ["ma.Schema"],
                "generate": True,
            },
            "invenio-ext-extra-code": "",
            "config-resource-class-key": "TEST_RESOURCE_CLASS_TEST",
            "record-mapping-setup-cfg": "test",
            "config-package": "test.config",
            "service-id": "test",
            "mapping-package": "test.records.mappings",
            "record-search-options-class": "test.services.records.search.TestSearchOptions",
            "config-resource-config-key": "TEST_RESOURCE_CONFIG_TEST",
            "invenio-record-permissions-extra-code": "",
            "invenio-record-resource-extra-code": "",
            "index-name": "test-test-1.0.0",
            "generate-record-pid-field": True,
            "record-jsonschemas-setup-cfg": "test",
            "record-facets-class": "test.services.records.facets.Test",
            "create-blueprint-from-app": "test.views.create_blueprint_from_app_test",
            "schema-server": "http://localhost/schemas/",
            "record-schema-metadata-setup-cfg": "test",
            "record-permissions-class": "test.services.records.permissions.TestPermissionPolicy",
            "mapping-file": "test/records/mappings/os-v2/test/test-1.0.0.json",
            "record-schema-class": "test.services.records.schema.TestSchema",
            "package-path": "test",
            "record-prefix-snake": "test",
            "script-import-sample-data": "data/sample_data.yaml",
            "ext-class": "test.ext.TestExt",
            "record-service-class": "test.services.records.service.TestService",
            "record-service-config-class": "test.services.records.config.TestServiceConfig",
            "record-dumper-extensions": [],
            "record-resource-blueprint-name": "Test",
            "config-service-class-key": "TEST_SERVICE_CLASS_TEST",
            "invenio-record-resource-config-extra-code": "",
            "jsonschemas-package": "test.records.jsonschemas",
            "package": "test",
            "extension-suffix": "test",
            "saved-model-file": "test/models/model.json",
            "record-schema-metadata-alembic": "test",
            "invenio-record-facets-extra-code": "",
            "record-class": "test.records.api.TestRecord",
            "properties": {
                "a": {
                    "marshmallow": {
                        "field-class": "ma_fields.String",  # NOSONAR
                        "validators": [],
                        "imports": [],
                    },
                    "type": "keyword",
                    "ui": {"class": "bolder"},
                },
                "b": {
                    "marshmallow": {
                        "generate": True,
                        "field-class": "ma_fields.Nested",
                        "validators": [],
                        "schema-class": "test.services.records.schema.BSchema",
                        "imports": [],
                    },
                    "type": "object",
                    "properties": {
                        "c": {
                            "marshmallow": {
                                "field-class": "ma_fields.String",
                                "validators": [],
                                "imports": [],
                            },
                            "type": "keyword",
                        }
                    },
                },
                "metadata": {
                    "marshmallow": {
                        "schema-class": "test.services.records.schema.TestMetadataSchema",  # NOSONAR
                        "base-classes": ["ma.Schema"],
                        "generate": True,
                        "field-class": "ma_fields.Nested",
                        "validators": [],
                        "imports": [],
                    },
                    "type": "object",
                    "properties": {},
                },
            },
            "flask-commands-setup-cfg": "test",
            "package-base": "test",
            "record-resources-package": "test.resources.records",
            "oarepo-models-setup-cfg": "test",
            "package-base-upper": "TEST",
            "invenio-record-extra-code": "",
            "config-resource-register-blueprint-key": "TEST_REGISTER_BLUEPRINT",
            "flask-extension-name": "test",
            "invenio-views-extra-code": "",
            "record-records-package": "test.records",
            "record-blueprints-setup-cfg": "test",
            "record-service-config-generate-links": True,
            "invenio-version-extra-code": "",
            "record-metadata-table-name": "test_metadata",
            "invenio-config-extra-code": "",
            "cli-function": "test.cli.group",
            "profile-package": "records",
            "schema-name": "test-1.0.0.json",
            "collection-url": "/test/",
            "invenio-record-search-options-extra-code": "",
            "proxies-current-resource": "test.proxies.current_resource",
            "invenio-proxies-extra-code": "",
            "config-service-config-key": "TEST_SERVICE_CONFIG_TEST",
            "record-api-blueprints-setup-cfg": "test",
            "invenio-record-service-config-extra-code": "",
            "record-resource-config-class": "test.resources.records.config.TestResourceConfig",
            "invenio-record-object-schema-extra-code": "",
            "record-metadata-class": "test.records.models.TestMetadata",
            "schema-file": "test/records/jsonschemas/test-1.0.0.json",
            "config-dummy-class": "test.config.DummyClass",
            "kebap-package": "test",
            "model-name": "test",
            "invenio-record-schema-extra-code": "",
            "invenio-record-service-extra-code": "",
            "invenio-record-metadata-extra-code": "",
            "record-schema-metadata-class": "test.services.records.schema.TestMetadataSchema",
            "record-prefix": "Test",
            "proxies-current-service": "test.proxies.current_service",
            "record-services-package": "test.services.records",
        }
    }
    assert data[1].strip() == ""
    assert (
        data[2].strip()
        == """[options.entry_points]
oarepo.models = test = test.models:model.json"""
    )


def build(model, output_builder_components=None, property_preprocessors=None):
    datatypes._prepare_datatypes()
    if UIValidator not in model_validator.validator_map["property"]:
        model_validator.validator_map["property"].append(UIValidator)
    datatypes.datatype_map["multilingual"] = MultilingualDataType

    builder = ModelBuilder(
        output_builders=[
            ModelSaverBuilder,
            ModelRegistrationBuilder,
        ],
        outputs=[JSONOutput, PythonOutput, CFGOutput],
        model_preprocessors=[DefaultValuesModelPreprocessor, InvenioModelPreprocessor],
        output_builder_components=output_builder_components,
        filesystem=InMemoryFileSystem(),
        property_preprocessors=property_preprocessors,
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {
                        "use-isort": False,
                        "use-black": False,
                    },
                },
                "model": {
                    **model,
                    "package": "test",
                    "record-prefix": "Test",
                },
            },
            included_models=load_included_models_from_entry_points(),
            loaders=load_entry_points_dict("oarepo_model_builder.loaders"),
        ),
        output_dir="",
    )
    return (
        json5.load(
            builder.filesystem.open(os.path.join("test", "models", "model.json"))
        ),
        builder.filesystem.read(os.path.join("test", "models", "__init__.py")),
        builder.filesystem.read("setup.cfg"),
    )


def test_model_saver_invenio():
    data = build(
        {"use": "invenio", "properties": {}},
        property_preprocessors=[
            DataTypePreprocessor,
        ],
    )
    print(json.dumps(data[0], indent=4))
    assert data[0] == {
        "model": {
            "type": "object",
            "config-dummy-class": "test.config.DummyClass",
            "record-resources-package": "test.resources.records",
            "package-base-upper": "TEST",
            "record-class": "test.records.api.TestRecord",
            "record-resource-class": "test.resources.records.resource.TestResource",
            "invenio-views-extra-code": "",
            "record-resource-config-class": "test.resources.records.config.TestResourceConfig",
            "marshmallow": {
                "schema-class": "test.services.records.schema.TestSchema",
                "generate": True,
                "base-classes": ["InvenioBaseRecordSchema"],
            },
            "package": "test",
            "record-blueprints-setup-cfg": "test",
            "record-schema-class": "test.services.records.schema.TestSchema",
            "record-prefix-snake": "test",
            "record-records-package": "test.records",
            "ext-class": "test.ext.TestExt",
            "invenio-config-extra-code": "",
            "record-services-package": "test.services.records",
            "properties": {
                "id": {
                    "marshmallow": {
                        "write": False,
                        "read": False,
                        "field-class": "ma_fields.String",
                        "validators": [],
                        "imports": [],
                    },
                    "type": "keyword",
                    "sample": {"skip": True},
                },
                "created": {
                    "marshmallow": {
                        "write": False,
                        "read": True,
                        "field-class": "ma_fields.String",
                        "validators": ["validate_date('%Y:%m:%d')"],
                        "imports": [
                            {"import": "oarepo_runtime.validation.validate_date"}
                        ],
                    },
                    "type": "date",
                    "sample": {"skip": True},
                },
                "updated": {
                    "marshmallow": {
                        "write": False,
                        "read": True,
                        "field-class": "ma_fields.String",
                        "validators": ["validate_date('%Y:%m:%d')"],
                        "imports": [
                            {"import": "oarepo_runtime.validation.validate_date"}
                        ],
                    },
                    "type": "date",
                    "sample": {"skip": True},
                },
                "$schema": {
                    "marshmallow": {
                        "write": False,
                        "read": False,
                        "field-class": "ma_fields.String",
                        "validators": [],
                        "imports": [],
                    },
                    "type": "keyword",
                    "sample": {"skip": True},
                },
            },
            "config-resource-register-blueprint-key": "TEST_REGISTER_BLUEPRINT",
            "record-facets-class": "test.services.records.facets.Test",
            "record-permissions-class": "test.services.records.permissions.TestPermissionPolicy",
            "create-blueprint-from-app": "test.views.create_blueprint_from_app_test",
            "script-import-sample-data": "data/sample_data.yaml",
            "record-service-config-class": "test.services.records.config.TestServiceConfig",
            "invenio-record-extra-code": "",
            "record-dumper-class": "test.records.dumper.TestDumper",
            "jsonschemas-package": "test.records.jsonschemas",
            "record-resource-blueprint-name": "Test",
            "config-resource-class-key": "TEST_RESOURCE_CLASS_TEST",
            "invenio-record-dumper-extra-code": "",
            "record-dumper-extensions": [],
            "collection-url": "/test/",
            "schema-server": "http://localhost/schemas/",
            "service-id": "test",
            "record-search-options-class": "test.services.records.search.TestSearchOptions",
            "mapping-package": "test.records.mappings",
            "invenio-record-permissions-extra-code": "",
            "model-name": "test",
            "oarepo-models-setup-cfg": "test",
            "invenio-record-object-schema-extra-code": "",
            "schema-name": "test-1.0.0.json",
            "index-name": "test-test-1.0.0",
            "mapping-file": "test/records/mappings/os-v2/test/test-1.0.0.json",
            "invenio-version-extra-code": "",
            "package-path": "test",
            "config-service-class-key": "TEST_SERVICE_CLASS_TEST",
            "proxies-current-service": "test.proxies.current_service",
            "invenio-record-resource-config-extra-code": "",
            "package-base": "test",
            "invenio-record-search-options-extra-code": "",
            "invenio-record-service-extra-code": "",
            "invenio-record-resource-extra-code": "",
            "invenio-record-facets-extra-code": "",
            "flask-extension-name": "test",
            "invenio-record-service-config-extra-code": "",
            "saved-model-file": "test/models/model.json",
            "record-schema-metadata-alembic": "test",
            "record-schema-metadata-setup-cfg": "test",
            "invenio-ext-extra-code": "",
            "schema-file": "test/records/jsonschemas/test-1.0.0.json",
            "kebap-package": "test",
            "proxies-current-resource": "test.proxies.current_resource",
            "cli-function": "test.cli.group",
            "record-mapping-setup-cfg": "test",
            "config-package": "test.config",
            "record-service-config-generate-links": True,
            "invenio-record-metadata-extra-code": "",
            "record-prefix": "Test",
            "generate-record-pid-field": True,
            "extension-suffix": "test",
            "record-schema-metadata-class": "test.services.records.schema.TestMetadataSchema",
            "record-jsonschemas-setup-cfg": "test",
            "record-metadata-table-name": "test_metadata",
            "invenio-proxies-extra-code": "",
            "record-metadata-class": "test.records.models.TestMetadata",
            "schema-version": "1.0.0",
            "profile-package": "records",
            "record-api-blueprints-setup-cfg": "test",
            "record-service-class": "test.services.records.service.TestService",
            "config-resource-config-key": "TEST_RESOURCE_CONFIG_TEST",
            "config-service-config-key": "TEST_SERVICE_CONFIG_TEST",
            "invenio-record-schema-extra-code": "",
            "flask-commands-setup-cfg": "test",
        }
    }



from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch

from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_service"
    class_config = "record-service-class"
    template = "record-service"



from oarepo_model_builder.builders.extend import ExtendBuilder
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import \
    InvenioRecordSchemaBuilder
from oarepo_model_builder.invenio.invenio_script_sample_data import \
    InvenioScriptSampleDataBuilder
from oarepo_model_builder.property_preprocessors import (PropertyPreprocessor,
                                                         process)
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class DisableMarshmallowPreprocessor(PropertyPreprocessor):
    TYPE = "disable-marshmallow"

    @process(
        model_builder=ExtendBuilder,
        path="/properties/**",
        condition=lambda current, stack: stack.top.schema_element_type == "property",
    )
    def modify_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        marshmallow_def = "marshmallow"
        if marshmallow_def not in data:
            data["marshmallow"] = {"read": False, "write": False}
            return data
        marshmallow = data[marshmallow_def]
        if "schema-class" in marshmallow:
            # already generated - if user wants to override this, he has to set schema-class and generate
            marshmallow["generate"] = False
            schema_class = marshmallow.pop("schema-class")
            if not marshmallow.get("field", None):
                marshmallow["inherited-schema-class"] = schema_class

        marshmallow["read"] = False
        marshmallow["write"] = False

        return data



import json
import os
import re
from pathlib import Path

import yaml

from oarepo_model_builder.entrypoints import (create_builder_from_entrypoints,
                                              load_model)
from oarepo_model_builder.fs import InMemoryFileSystem
from tests.utils import assert_python_equals


def test_array_shortcuts():
    schema = load_model(
        "test.yaml",  # NOSONAR
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {"a[]": {"type": "keyword", "^required": True}},
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
from marshmallow import ValidationError
from marshmallow import validates as ma_validates
import marshmallow as ma
from marshmallow import fields as ma_fields
from marshmallow_utils import fields as mu_fields
from marshmallow_utils import schemas as mu_schemas

from oarepo_runtime.validation import validate_date

class TestSchema(InvenioBaseRecordSchema):
    \"""TestSchema schema.\"""
    a = ma_fields.List(ma_fields.String())        
    created = ma_fields.String(validate=[validate_date('%Y:%m:%d')], dump_only=True)
    updated = ma_fields.String(validate=[validate_date('%Y:%m:%d')], dump_only=True)
    """,
    )

    data = builder.filesystem.read(
        os.path.join("test", "records", "mappings", "os-v2", "test", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "mappings": {
            "properties": {
                "$schema": {"type": "keyword"},
                "a": {
                    "type": "keyword",
                },
                "created": {"type": "date"},
                "id": {"type": "keyword"},
                "updated": {"type": "date"},
            }
        }
    }

    data = builder.filesystem.read(
        os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "properties": {
            "$schema": {"type": "string"},
            "a": {"items": {"type": "string"}, "type": "array"},
            "created": {"format": "date", "type": "string"},
            "id": {"type": "string"},
            "updated": {"format": "date", "type": "string"},
        },
        "required": ["a"],
        "type": "object",
    }


def test_singleline_type_shortcut():
    data = """
model:
  properties:
    metadata:
      properties:
        title: fulltext
"""

    schema = load_model(
        "test.yaml",
        "test",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")


def test_object_shortcut():
    data = """
model:
  properties:
    metadata{}:
      title: fulltext
"""

    schema = load_model(
        "test.yaml",
        "test",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")


def test_array_shortcut():
    data = """
model:
  properties:
    metadata[]:
      type: fulltext
"""

    schema = load_model(
        "test.yaml",
        "test",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")


def test_array_single_line_shortcut():
    data = """
model:
  properties:
    metadata[]: fulltext
"""

    schema = load_model(
        "test.yaml",
        "test",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")


def test_misc_shortcuts():
    data = """
model:
  properties:
    metadata:
      properties:
        title: fulltext
"""
    """
      description: fulltext+keyword
      kind: keyword{enum:[article, dataset]}
      printed{}:
      ^marshmallow:
        schema-class: PrintedPublicationSchema
      price: double{minimumExclusive:0}
      floatPrice: float     # just to test floats
      published: date
      "sources{nested}[]":
        source: fulltext
        period: edtf-interval
"""

    schema = load_model(
        "test.yaml",
        "test",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")



import sys

import libcst as cst
from jinja2 import Environment, FunctionLoader, pass_context

from oarepo_model_builder.outputs import OutputBase
from oarepo_model_builder.templates import templates
from oarepo_model_builder.utils.cst import PythonContext, merge
from oarepo_model_builder.utils.jinja import (base_name, in_different_package,
                                              package_name,
                                              with_defined_prefix)
from oarepo_model_builder.utils.verbose import log


class PythonOutput(OutputBase):
    TYPE = "python"
    cst = None
    original_data = None

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = f.read()

                self.cst = cst.parse_module(self.original_data)
        except FileNotFoundError:
            self.original_data = None
            self.cst = cst.parse_module("")

    @property
    def created(self):
        return self.original_data is None

    def finish(self):
        code = self.cst.code
        if code != self.original_data:
            self.builder.filesystem.mkdir(self.path.parent)
            log(2, "Saving %s", self.path)
            with self.builder.filesystem.open(self.path, mode="w") as f:
                f.write(code)
            if self.builder.schema.settings.python.use_isort:
                import isort

                config = isort.settings.Config(verbose=False, quiet=True)
                isort.file(self.path, config=config)
            if self.builder.schema.settings.python.use_black:
                import subprocess

                subprocess.call(["black", "-q", "--preview", str(self.path)])

    def merge(self, template_name, context, filters=None):
        # template is a loadable resource
        env = Environment(
            loader=FunctionLoader(
                lambda tn: templates.get_template(tn, context["settings"])
            ),
            autoescape=False,
        )
        self.register_default_filters(env)
        for filter_name, filter_func in (filters or {}).items():
            env.filters[filter_name] = filter_func

        try:
            context = {
                **context,
                **{k.replace("-", "_"): v for k, v in list(context.items())},
            }
            rendered = env.get_template(template_name).render(context)
        except Exception as exc:
            raise RuntimeError(f"Error rendering template {template_name}") from exc
        try:
            rendered_cst = cst.parse_module(
                rendered, config=self.cst.config_for_parsing
            )
        except:
            print(rendered, file=sys.stderr)
            raise
        self.cst = merge(
            PythonContext(
                rendered_cst, conflict_resolver=self.builder.conflict_resolver
            ),
            self.cst,
            rendered_cst,
        )

    @staticmethod
    def register_default_filters(env):
        env.filters["package_name"] = package_name
        env.filters["base_name"] = pass_context(
            lambda context, value: base_name(value)
            if not with_defined_prefix(
                context["settings"].python.always_defined_import_prefixes, value
            )
            else value
        )
        env.tests["in_different_package"] = pass_context(
            lambda context, value: in_different_package(
                context["current_package_name"], value
            )
        )
        env.tests["not_prefixed"] = pass_context(
            lambda context, value: not with_defined_prefix(
                context["settings"].python.always_defined_import_prefixes, value
            )
        )
